from pyletteyes import Colour
from pyletteyes import Palette

# Create a colour and print RGB string
red = Colour.from_hex("#FF0000")
print(red)

# Lighten it and output as HSL values
light_red = red.lighten(0.1)
print(light_red.hsl)

# Darken it and print as an RGB string
dark_red = red.darken(0.2)
print(dark_red)

# Get complementary colour and output as hex
complement = red.get_complementary()
print(complement.to_hex())

# make a pastel of the complementary colour
pastel = complement.get_pastel()
print(pastel)

# make a palette of colours
palette = Palette([pastel, dark_red, red])

# output as a list of rgb strings
string_list = palette.to_string_list()
print(string_list)

# or a list of hex codes
hex_list = palette.to_hex_list()
print(hex_list)

# import as a list of rgb strings
from_string_list = Palette.from_string_list(string_list)
print(from_string_list.to_string_list())

# import as a list of hex strings
from_hex_list = Palette.from_hex_list(hex_list)
print(from_hex_list.to_hex_list())




import pytest
import numpy as np
from pyletteyes.palette import Palette
from pyletteyes.colour import Colour


@pytest.fixture
def basic_palette():
    """Create a basic palette with three colours."""
    return Palette([
        Colour(255, 0, 0),  # Red
        Colour(0, 255, 0),  # Green
        Colour(0, 0, 255)  # Blue
    ])


@pytest.fixture
def monochrome_palette():
    """Create a monochrome palette with three shades of blue."""
    return Palette([
        Colour(0, 0, 255),  # Blue
        Colour(0, 0, 192),  # Darker blue
        Colour(0, 0, 128)  # Even darker blue
    ])


def test_palette_initialization():
    # Test valid initialization
    colours = [Colour(255, 0, 0), Colour(0, 255, 0)]
    palette = Palette(colours)
    assert len(palette) == 2

    # Test empty palette
    with pytest.raises(ValueError):
        Palette([])

    # Test that palette makes a copy of the input list
    colours.append(Colour(0, 0, 255))
    assert len(palette) == 2


def test_palette_properties(basic_palette):
    assert basic_palette.size == 3
    assert len(basic_palette.colours) == 3

    # Test that colours property returns a copy
    colours = basic_palette.colours
    colours.append(Colour(0, 0, 0))
    assert len(basic_palette) == 3


def test_add_remove_colour(basic_palette):
    new_colour = Colour(255, 255, 0)  # Yellow
    initial_size = len(basic_palette)

    # Test adding
    basic_palette.add_colour(new_colour)
    assert len(basic_palette) == initial_size + 1
    assert new_colour in basic_palette.colours

    # Test removing
    basic_palette.remove_colour(new_colour)
    assert len(basic_palette) == initial_size
    assert new_colour not in basic_palette.colours

    # Test removing non-existent colour
    with pytest.raises(ValueError):
        basic_palette.remove_colour(Colour(128, 128, 128))

    # Test removing last colour
    while len(basic_palette) > 1:
        basic_palette.remove_colour(basic_palette[0])
    with pytest.raises(ValueError):
        basic_palette.remove_colour(basic_palette[0])


def test_hex_conversion():
    hex_colours = ["#FF0000", "#00FF00", "#0000FF"]
    palette = Palette.from_hex_list(hex_colours)

    assert len(palette) == 3
    assert palette.to_hex_list() == hex_colours


def test_contrast_score(basic_palette, monochrome_palette):
    # High contrast palette should have higher score
    high_contrast = basic_palette.score_contrast()

    # Low contrast (monochrome) palette should have lower score
    low_contrast = monochrome_palette.score_contrast()

    assert 0 <= high_contrast <= 1
    assert 0 <= low_contrast <= 1
    assert high_contrast > low_contrast


def test_harmony_score(basic_palette, monochrome_palette):
    # Test that scores are in valid range
    assert 0 <= basic_palette.score_harmony() <= 1
    assert 0 <= monochrome_palette.score_harmony() <= 1

    # Monochrome palette should be more harmonious
    assert monochrome_palette.score_harmony() > basic_palette.score_harmony()


def test_brightness_score():
    # Test white palette
    white_palette = Palette([Colour(255, 255, 255)])
    assert pytest.approx(white_palette.score_brightness(), abs=0.01) == 1.0

    # Test black palette
    black_palette = Palette([Colour(0, 0, 0)])
    assert pytest.approx(black_palette.score_brightness(), abs=0.01) == 0.0

    # Test mixed palette
    mixed_palette = Palette([
        Colour(255, 255, 255),  # White
        Colour(0, 0, 0)  # Black
    ])
    assert pytest.approx(mixed_palette.score_brightness(), abs=0.01) == 0.5


def test_dominant_colour(basic_palette):
    dominant = basic_palette.get_dominant_colour()
    assert isinstance(dominant, Colour)
    assert dominant == basic_palette[0]


def test_iteration(basic_palette):
    # Test that palette is iterable
    colours = list(basic_palette)
    assert len(colours) == 3
    assert all(isinstance(c, Colour) for c in colours)


def test_indexing(basic_palette):
    # Test positive indexing
    assert isinstance(basic_palette[0], Colour)
    assert basic_palette[0].rgb == (255, 0, 0)

    # Test negative indexing
    assert basic_palette[-1].rgb == (0, 0, 255)

    # Test index out of range
    with pytest.raises(IndexError):
        _ = basic_palette[len(basic_palette)]


def test_representation(basic_palette):
    repr_str = repr(basic_palette)
    assert repr_str.startswith("Palette(colours=[")
    for colour in basic_palette.colours:
        assert colour.to_hex() in repr_str
    assert repr_str.endswith("])")


def test_uniqueness_score(basic_palette, monochrome_palette):
    # Test that scores are in valid range
    assert 0 <= basic_palette.score_uniqueness() <= 1
    assert 0 <= monochrome_palette.score_uniqueness() <= 1

    # Basic palette (Red, Green, Blue) should have high uniqueness
    high_uniqueness = basic_palette.score_uniqueness()

    # Monochrome palette should have low uniqueness
    low_uniqueness = monochrome_palette.score_uniqueness()

    assert high_uniqueness > low_uniqueness
    assert high_uniqueness > 0.8  # RGB colours should be very unique
    assert low_uniqueness < 0.5  # Similar blues should have low uniqueness


def test_uniqueness_score_single_colour():
    # Single colour palette should have perfect uniqueness
    palette = Palette([Colour(255, 0, 0)])  # Single red
    assert pytest.approx(palette.score_uniqueness(), abs=0.01) == 1.0


def test_uniqueness_score_identical_colours():
    # Palette with identical colours should have minimum uniqueness
    palette = Palette([
        Colour(255, 0, 0),  # Red
        Colour(255, 0, 0)  # Same red
    ])
    assert pytest.approx(palette.score_uniqueness(), abs=0.01) == 0.0


def test_uniqueness_score_similar_colours():
    # Test with very similar but not identical colours
    palette = Palette([
        Colour(255, 0, 0),  # Red
        Colour(250, 5, 5)  # Very similar red
    ])
    score = palette.score_uniqueness()
    assert score < 0.2  # Should have low uniqueness score


def test_brightness_balance_score_perfect_balance():
    palette = Palette([
        Colour.from_hsl((0, 0, 0.3)),
        Colour.from_hsl((0, 0, 0.7))
    ])
    assert pytest.approx(palette.score_brightness_balance(), abs=0.01) == 1


def test_brightness_balance_score_all_dark():
    palette = Palette([
        Colour.from_hsl((0, 0, 0.1)),
        Colour.from_hsl((0, 0, 0.2)),
        Colour.from_hsl((0, 0, 0.1))
    ])

    # calculate the mean lightness as an expected low score
    mean_lightness = np.mean([0.1, 0.2, 0.1])

    expected = 1 - abs(mean_lightness - 0.5) * 2
    assert pytest.approx(palette.score_brightness_balance(), abs(0.01)) == expected


def test_brightness_balance_score_all_light():
    palette = Palette([
        Colour.from_hsl((0, 0, 0.8)),
        Colour.from_hsl((0, 0, 0.9)),
        Colour.from_hsl((0, 0, 0.8))
    ])

    # calculate the mean lightness as an expected low score
    mean_lightness = np.mean([0.8, 0.9, 0.8])

    expected = 1 - abs(mean_lightness - 0.5) * 2
    assert pytest.approx(palette.score_brightness_balance(), abs(0.01)) == expected


def test_brightness_balance_score_mixed_colours():
    palette = Palette([
        Colour.from_hsl((0, 0, 0.2)),
        Colour.from_hsl((0, 0, 0.4)),
        Colour.from_hsl((0, 0, 0.6)),
        Colour.from_hsl((0, 0, 0.8))
    ])

    # calculate the mean lightness as an expected low score
    mean_lightness = np.mean([0.2, 0.4, 0.6, 0.8])

    expected = 1 - abs(mean_lightness - 0.5) * 2
    assert pytest.approx(palette.score_brightness_balance(), abs(0.01)) == expected


def test_from_string_list():
    string_palette = ["rgb(255, 0, 0)", "rgb(0, 255, 0)", "rgb(0, 0, 255)"]
    converted_palette = Palette.from_string_list(string_palette)
    expected_palette = Palette([Colour(255,0,0), Colour(0,255,0), Colour(0,0,255)])

    assert converted_palette.colours == expected_palette.colours


def test_to_string_list():
    expected_palette = ["rgb(255, 0, 0)", "rgb(0, 255, 0)", "rgb(0, 0, 255)"]
    input_palette = Palette([Colour(255, 0, 0), Colour(0, 255, 0), Colour(0, 0, 255)])

    assert input_palette.to_string_list() == expected_palette


def test_single_colour_palette():
    scp = Palette.from_hex_list(["#FF0000"])

    assert scp.score_brightness_balance() == 1.0
    assert scp.score_temperature_variation() == 1.0
    assert scp.score_saturation_variation() == 1.0
    assert scp.score_harmony() == 1.0
    assert scp.score_contrast() == 1.0


import pytest
from unittest.mock import patch
from pyletteyes.colour import Colour


def test_colour_initialization():
    # Test valid initialization
    c = Colour(255, 128, 0)
    assert c.rgb == (255, 128, 0)

    # Test float inputs (should round)
    c = Colour(255.4, 128.6, 0.2)
    assert c.rgb == (255, 129, 0)

    # Test invalid values
    with pytest.raises(ValueError):
        Colour(256, 128, 0)
    with pytest.raises(ValueError):
        Colour(-1, 128, 0)
    with pytest.raises(ValueError):
        Colour(255, 128, 300)


def test_hex_conversion():
    # Test hex to RGB
    c = Colour.from_hex("#FF8000")
    assert c.rgb == (255, 128, 0)

    # Test RGB to hex
    c = Colour(255, 128, 0)
    assert c.to_hex().upper() == "#FF8000"

    # Test hex without #
    c = Colour.from_hex("FF8000")
    assert c.rgb == (255, 128, 0)

    # Test invalid hex
    with pytest.raises(ValueError):
        Colour.from_hex("#FF80")  # Too short
    with pytest.raises(ValueError):
        Colour.from_hex("#GG8000")  # Invalid characters


def test_hsl_conversion():
    # Test pure red
    c = Colour(255, 0, 0)
    h, s, l = c.hsl
    assert pytest.approx(h, abs=0.01) == 0.0  # Red hue
    assert pytest.approx(s, abs=0.01) == 1.0  # Full saturation
    assert pytest.approx(l, abs=0.01) == 0.5  # Mid lightness


def test_from_hsl_valid():
    """Test valid HSL to RGB conversion."""
    test_cases = [
        ((0.0, 0.0, 0.0), (0, 0, 0)),  # Black
        ((0.0, 0.0, 1.0), (255, 255, 255)),  # White
        ((0.0, 1.0, 0.5), (255, 0, 0)),  # Pure red
        ((1 / 3, 1.0, 0.5), (0, 255, 0)),  # Pure green (exactly 1/3)
        ((2 / 3, 1.0, 0.5), (0, 0, 255))  # Pure blue (exactly 2/3) # Blue
    ]

    for hsl, expected_rgb in test_cases:
        colour = Colour.from_hsl(hsl)
        assert (colour._r, colour._g, colour._b) == expected_rgb


def test_from_hsl_out_of_range():
    """Test HSL values outside the valid range."""
    invalid_inputs = [
        (-0.1, 0.5, 0.5),  # Negative hue
        (0.5, -0.1, 0.5),  # Negative saturation
        (0.5, 0.5, -0.1),  # Negative lightness
        (1.1, 0.5, 0.5),  # Hue > 1
        (0.5, 1.1, 0.5),  # Saturation > 1
        (0.5, 0.5, 1.1)  # Lightness > 1
    ]

    for hsl in invalid_inputs:
        with pytest.raises(ValueError, match="HSL values must be within the range of 0.0 to 1.0"):
            Colour.from_hsl(hsl)


# def test_from_hsl_colorsys_error():
#     """Test handling of ValueError from colorsys.hls_to_rgb."""
#     with patch('colorsys.hls_to_rgb') as mock_hls_to_rgb:
#         mock_hls_to_rgb.side_effect = ValueError("Mock colorsys error")
#
#         with pytest.raises(ValueError, match="Invalid HSL value"):
#             Colour.from_hsl((0.5, 0.5, 0.5))


def test_lighten_darken():
    c = Colour(100, 100, 100)

    # Test lightening
    lighter = c.lighten(0.2)
    assert all(x > y for x, y in zip(lighter.rgb, c.rgb))

    # Test darkening
    darker = c.darken(0.2)
    assert all(x < y for x, y in zip(darker.rgb, c.rgb))

    # Test maximum lightness
    white = Colour(255, 255, 255)
    assert white.lighten(0.1).rgb == (255, 255, 255)

    # Test minimum darkness
    black = Colour(0, 0, 0)
    assert black.darken(0.1).rgb == (0, 0, 0)


def test_complementary():
    # Test red -> cyan
    red = Colour(255, 0, 0)
    cyan = red.get_complementary()
    assert pytest.approx(cyan.rgb, abs=1) == (0, 255, 255)

    # Test green -> magenta
    green = Colour(0, 255, 0)
    magenta = green.get_complementary()
    assert pytest.approx(magenta.rgb, abs=1) == (255, 0, 255)


def test_analogous():
    # Test red's analogous colours
    red = Colour(255, 0, 0)
    left, right = red.get_analogous()

    # Check that the analogous colours are different from the original
    assert left.rgb != red.rgb
    assert right.rgb != red.rgb

    # Check that the analogous colours are different from each other
    assert left.rgb != right.rgb

    # Test custom angle
    left, right = red.get_analogous(angle=15)
    left2, right2 = red.get_analogous(angle=45)

    # Wider angle should produce more different colours
    assert sum(abs(a - b) for a, b in zip(left.rgb, right.rgb)) < \
           sum(abs(a - b) for a, b in zip(left2.rgb, right2.rgb))


def test_equality():
    c1 = Colour(255, 128, 0)
    c2 = Colour(255, 128, 0)
    c3 = Colour(0, 128, 255)

    assert c1 == c2
    assert c1 != c3
    assert c1 != "not a colour"


def test_string_representation():
    c = Colour(255, 128, 0)
    assert repr(c) == "rgb(255, 128, 0)"


# Parameterized tests for various colour values
@pytest.mark.parametrize("r,g,b,hex_code", [
    (255, 0, 0, "#FF0000"),
    (0, 255, 0, "#00FF00"),
    (0, 0, 255, "#0000FF"),
    (0, 0, 0, "#000000"),
    (255, 255, 255, "#FFFFFF"),
    (128, 128, 128, "#808080"),
])
def test_colour_values(r, g, b, hex_code):
    c = Colour(r, g, b)
    assert c.to_hex().upper() == hex_code
    assert Colour.from_hex(hex_code).rgb == (r, g, b)


def test_valid_from_string():
    """Test parsing of valid RGB strings."""
    test_cases = [
        ("rgb(255, 0, 0)", (255, 0, 0)),
        ("rgb(0, 255, 0)", (0, 255, 0)),
        ("rgb(0, 0, 255)", (0, 0, 255)),
        ("rgb(128, 128, 128)", (128, 128, 128)),
        ("rgb(0, 0, 0)", (0, 0, 0)),
        ("rgb(255, 255, 255)", (255, 255, 255))
    ]

    for rgb_string, expected in test_cases:
        colour = Colour.from_string(rgb_string)
        assert colour._r == expected[0]
        assert colour._g == expected[1]
        assert colour._b == expected[2]


def test_invalid_from_string():
    """Test that invalid RGB strings raise appropriate errors."""
    invalid_inputs = [
        "rgb(256, 0, 0)",  # Red value too high
        "rgb(-1, 0, 0)",  # Negative value
        "rgb(0, 0",  # Missing closing parenthesis
        "rgb(0, 0, 0, 0)",  # Too many values
        "rgb(a, b, c)",  # Non-numeric values
        "rgb()",  # Empty string
        "rgb(0, 0,)",  # Missing value
        "rgb(255.5, 0, 0)"  # Decimal values
    ]

    for invalid_input in invalid_inputs:
        with pytest.raises(ValueError):
            Colour.from_string(invalid_input)


def test_to_string():
    """Test conversion of Colour instances to RGB strings."""
    test_cases = [
        ((255, 0, 0), "rgb(255, 0, 0)"),
        ((0, 255, 0), "rgb(0, 255, 0)"),
        ((0, 0, 255), "rgb(0, 0, 255)"),
        ((128, 128, 128), "rgb(128, 128, 128)"),
        ((0, 0, 0), "rgb(0, 0, 0)"),
        ((255, 255, 255), "rgb(255, 255, 255)")
    ]

    for rgb_values, expected in test_cases:
        colour = Colour(*rgb_values)
        assert colour.to_string() == expected


def test_roundtrip():
    """Test conversion from string to Colour and back preserves values."""
    test_strings = [
        "rgb(255, 0, 0)",
        "rgb(0, 255, 0)",
        "rgb(0, 0, 255)",
        "rgb(128, 128, 128)",
        "rgb(0, 0, 0)",
        "rgb(255, 255, 255)"
    ]

    for rgb_string in test_strings:
        colour = Colour.from_string(rgb_string)
        assert colour.to_string() == rgb_string


def test_edge_cases():
    """Test edge cases for RGB string parsing."""
    # Extra spaces shouldn't matter
    colour = Colour.from_string("rgb(255, 0, 0)")
    assert colour.to_string() == "rgb(255, 0, 0)"

    # Whitespace at start/end shouldn't matter
    colour = Colour.from_string("  rgb(255, 0, 0)  ")
    assert colour.to_string() == "rgb(255, 0, 0)"


from colorsys import rgb_to_hls, hls_to_rgb
from typing import Tuple, Union


class Colour:
    """A class representing a colour in RGB space with various transformation capabilities."""

    def __init__(self, r: Union[int, float], g: Union[int, float], b: Union[int, float]):
        """
        Initialize a new Colour instance.

        Args:
            r (int): Red component (0-255)
            g (int): Green component (0-255)
            b (int): Blue component (0-255)

        Raises:
            ValueError: If RGB values are not in valid range
        """
        rounded_r = round(r)
        rounded_g = round(g)
        rounded_b = round(b)

        if not all(isinstance(v, (int, float)) and 0 <= v <= 255 for v in (rounded_r, rounded_g, rounded_b)):
            raise ValueError("RGB values must be between 0 and 255")

        self._r = rounded_r
        self._g = rounded_g
        self._b = rounded_b

    @property
    def rgb(self) -> Tuple[int, int, int]:
        """Get the RGB values as a tuple."""
        return (self._r, self._g, self._b)

    @property
    def hsl(self) -> Tuple[float, float, float]:
        """Get the HSL values as a tuple (Hue: 0-1, Saturation: 0-1, Lightness: 0-1)."""
        hls = rgb_to_hls(self._r / 255, self._g / 255, self._b / 255)
        return hls[0], hls[2], hls[1]  # must rearrange HLS to HSL

    @classmethod
    def from_hsl(cls, hsl: Tuple[float, float, float]) -> 'Colour':
        """
        Create a Colour instance from HSL values.

        Args:
            HSL values (tuple): Hue, Saturation, Lightness (e.g., (0.1, 0.3, 0.5))

        Returns:
            Colour: New Colour instance

        Raises:
            ValueError: If HSL values are invalid
        """
        for i in hsl:
            if i < 0.0 or i > 1.0:
                raise ValueError("HSL values must be within the range of 0.0 to 1.0")

        try:
            r, g, b = hls_to_rgb(h=hsl[0], l=hsl[2], s=hsl[1])
            return cls(round(r * 255), round(g * 255), round(b * 255))
        except ValueError:
            raise ValueError("Invalid HSL value")

    @classmethod
    def from_string(cls, rgb_string: str) -> 'Colour':
        """
        Create a Colour instance from an RGB string.

        Args:
            rgb_string (str): RGB in string format (e.g., "rgb(255, 0, 0)")

        Returns:
            Colour: New Colour instance

        Raises:
            ValueError: If RGB string is invalid
        """
        rgb_string = rgb_string.replace('rgb(', '').replace(')', '').split(',')
        rgb_string = [int(i.strip()) for i in rgb_string]

        if len(rgb_string) != 3:
            raise ValueError("Invalid rgb string")

        try:
            r, g, b = rgb_string[0], rgb_string[1], rgb_string[2]
            return cls(r, g, b)
        except ValueError:
            raise ValueError("Invalid rgb string")


    def to_string(self) -> str:
        """Convert the colour to rgb string."""
        return f"rgb({self._r}, {self._g}, {self._b})"


    @classmethod
    def from_hex(cls, hex_string: str) -> 'Colour':
        """
        Create a Colour instance from a hex string.

        Args:
            hex_string (str): Hex colour code (e.g., "#FF0000" or "FF0000")

        Returns:
            Colour: New Colour instance

        Raises:
            ValueError: If hex string is invalid
        """
        hex_string = hex_string.lstrip('#')
        if len(hex_string) != 6:
            raise ValueError("Hex colour must be 6 characters long")

        try:
            r, g, b = tuple(int(hex_string[i:i + 2], 16) for i in (0, 2, 4))
            return cls(r, g, b)
        except ValueError:
            raise ValueError("Invalid hex colour string")

    def to_hex(self) -> str:
        """Convert the colour to hex format."""
        return f"#{self._r:02x}{self._g:02x}{self._b:02x}".upper()

    def lighten(self, amount: float = 0.1) -> 'Colour':
        """
        Create a lighter version of the colour.

        Args:
            amount (float): Amount to lighten by (0-1)

        Returns:
            Colour: New lightened Colour instance
        """
        h, s, l = self.hsl
        r, g, b = hls_to_rgb(h, min(1, l + amount), s)
        return Colour(int(r * 255), int(g * 255), int(b * 255))

    def darken(self, amount: float = 0.1) -> 'Colour':
        """
        Create a darker version of the colour.

        Args:
            amount (float): Amount to darken by (0-1)

        Returns:
            Colour: New darkened Colour instance
        """
        h, s, l = self.hsl
        r, g, b = hls_to_rgb(h, max(0, l - amount), s)
        return Colour(int(r * 255), int(g * 255), int(b * 255))

    def get_pastel(self) -> 'Colour':
        """
        Return a more 'pastel' version of the colour.
        Reduces saturation and increases brightness.

        :return:
            Colour: New pastel-ised Colour instance
        """
        h, s, l = self.hsl
        # Reduce saturation and increase lightness
        new_s = max(0.05, (s * 0.5))
        new_l = min(0.95, (l * 1.2))
        r, g, b = hls_to_rgb(h, new_l, new_s)

        return Colour(int(r * 255), int(g * 255), int(b * 255))

    def get_complementary(self) -> 'Colour':
        """
        Get the complementary colour (180 degrees opposite on the colour wheel).

        Returns:
            Colour: New Colour instance of the complementary colour
        """
        h, s, l = self.hsl
        # Add 0.5 to hue to get the opposite colour (180 degrees on the colour wheel)
        new_h = (h + 0.5) % 1.0
        r, g, b = hls_to_rgb(new_h, l, s)
        return Colour(int(r * 255), int(g * 255), int(b * 255))

    def get_analogous(self, angle: float = 30) -> Tuple['Colour', 'Colour']:
        """
        Get analogous colours (adjacent on the colour wheel).

        Args:
            angle (float): Angle of separation in degrees (default 30)

        Returns:
            Tuple[Colour, Colour]: Two new Colour instances
        """
        h, s, l = self.hsl
        angle = angle / 360  # Convert to 0-1 range

        h1 = (h + angle) % 1.0
        h2 = (h - angle) % 1.0

        r1, g1, b1 = hls_to_rgb(h1, l, s)
        r2, g2, b2 = hls_to_rgb(h2, l, s)

        return (
            Colour(int(r1 * 255), int(g1 * 255), int(b1 * 255)),
            Colour(int(r2 * 255), int(g2 * 255), int(b2 * 255))
        )

    def get_triadic(self) -> Tuple['Colour', 'Colour']:
        """
        Get triadic colours (shift hue left and right by 120 degrees).

        Returns:
            Tuple[Colour, Colour]: Two new Colour instances
        """
        return self.get_analogous(angle=120)


    def __eq__(self, other: object) -> bool:
        """Compare two colours for equality."""
        if not isinstance(other, Colour):
            return NotImplemented
        return self.rgb == other.rgb

    def __repr__(self) -> str:
        """String representation of the colour."""
        return f"rgb({self._r}, {self._g}, {self._b})"


from typing import List, Optional, Tuple
from .colour import Colour
import numpy as np


class Palette:
    """A class representing a collection of colours with analysis capabilities."""

    def __init__(self, colours: List[Colour]):
        """
        Initialize a new Palette instance.

        Args:
            colours (List[Colour]): List of Colour objects

        Raises:
            ValueError: If palette is empty
        """
        if not colours:
            raise ValueError("Palette must contain at least one colour")
        self._colours = colours.copy()  # Create a copy to prevent external modification

    @property
    def colours(self) -> List[Colour]:
        """Get the list of colours in the palette."""
        return self._colours.copy()  # Return a copy to prevent external modification

    @property
    def size(self) -> int:
        """Get the number of colours in the palette."""
        return len(self._colours)

    def score_contrast(self) -> float:
        """
        Calculate the overall contrast score of the palette.
        Higher scores indicate more contrast between colours.

        Returns:
            float: Contrast score between 0 and 1
        """
        if self.size < 2:
            return 1.0  # Single colour has no contrast

        contrasts = []
        for i, c1 in enumerate(self._colours):
            for c2 in self._colours[i + 1:]:
                # Calculate contrast using luminance difference
                # Convert RGB to luminance using standard coefficients
                l1 = 0.299 * c1.rgb[0] + 0.587 * c1.rgb[1] + 0.114 * c1.rgb[2]
                l2 = 0.299 * c2.rgb[0] + 0.587 * c2.rgb[1] + 0.114 * c2.rgb[2]

                # Normalize luminance difference
                contrast = abs(l1 - l2) / 255
                contrasts.append(contrast)

        return sum(contrasts) / len(contrasts)


    def score_uniqueness(self) -> float:
        """
        Score the palette based on how distinct colours are from one another.
        Higher scores indicate greater uniqueness.
        :return:
            float: Uniqueness score between 0 and 1
        """
        if self.size < 2:
            return 1.0

        rgb_values = np.array([
            [int(c.rgb[0]), int(c.rgb[1]), int(c.rgb[2])]
            for c in self._colours
        ], dtype=np.int32)

        similarities = []
        for i in range(len(rgb_values)):
            for j in range(i + 1, len(rgb_values)):
                # Max possible distance in RGB space is sqrt(255^2 * 3) ≈ 441.67
                distance = np.sqrt(np.sum((rgb_values[i].astype(float) - rgb_values[j].astype(float)) ** 2))
                similarity = 1 - (distance / 441.67)
                similarities.append(similarity)

        return float(1 - np.mean(similarities))


    def score_harmony(self) -> float:
        """
        Calculate the colour harmony score of the palette.
        Based on hue differences and standard colour theory principles.
        Higher scores indicate more harmonious combinations.

        Returns:
            float: Harmony score between 0 and 1
        """
        if self.size < 2:
            return 1.0

        # Get all HSL values
        hsls = [c.hsl for c in self._colours]

        # Calculate hue differences
        harmony_scores = []
        for i, (h1, _, _) in enumerate(hsls):
            for h2, _, _ in hsls[i + 1:]:
                # Calculate the smallest hue difference accounting for circular nature
                hue_diff = min(abs(h1 - h2), 1 - abs(h1 - h2))

                # Score based on common harmony principles
                # Perfect harmony at:
                # - 0 (same hue/monochromatic) = 1.0
                # - 1/6 (30° analogous) = 0.8
                # - 1/3 (120° triadic) = 0.9
                # - 1/2 (180° complementary) = 0.85
                harmony = max(
                    1.0 if hue_diff < 0.01 else 0,  # Monochromatic
                    0.8 if abs(hue_diff - 1 / 6) < 0.1 else 0,  # Analogous
                    0.9 if abs(hue_diff - 1 / 3) < 0.1 else 0,  # Triadic
                    0.85 if abs(hue_diff - 1 / 2) < 0.1 else 0,  # Complementary
                    # Base score for other relationships
                    0.5 - min(abs(hue_diff - 0), abs(hue_diff - 1 / 6),
                              abs(hue_diff - 1 / 3), abs(hue_diff - 1 / 2))
                )
                harmony_scores.append(harmony)

        return sum(harmony_scores) / len(harmony_scores)

    def score_saturation_variation(self) -> float:
        """
        Calculate a score based on variation in saturation levels.
        :return:
            float: Saturation variation score between 0 and 1
        """
        if self.size < 2:
            return 1.0

        saturations = [c.hsl[1] for c in self._colours]
        return float(np.std(saturations))


    def score_temperature_variation(self) -> float:
        """
        Calculate a score based on the mix of warm and cool colours.
        :return:
            float: Temperature variation score between 0 and 1
        """
        if self.size < 2:
            return 1.0

        hues = [c.hsl[0] for c in self._colours]
        warm_colors = sum(1 for h in hues if h <= 0.167 or h >= 0.833)
        cool_colors = len(hues) - warm_colors
        return float(1 - abs(warm_colors - cool_colors) / len(hues))


    def score_brightness(self) -> float:
        """
        Calculate the overall brightness score of the palette.

        Returns:
            float: Brightness score between 0 and 1
        """
        brightnesses = []
        for colour in self._colours:
            # Use the lightness component from HSL
            _, _, lightness = colour.hsl
            brightnesses.append(lightness)

        return sum(brightnesses) / len(brightnesses)


    def score_brightness_balance(self) -> float:
        """
        Calculate a score based on the balance of light and dark colours.
        :return:
            float: Brightness (lightness) balance score between 0 and 1
        """
        if self.size < 2:
            return 1.0

        values = [c.hsl[2] for c in self._colours]

        return float(1 - abs(np.mean(values) - 0.5) * 2)


    def get_dominant_colour(self) -> Colour:
        """
        Get the most dominant colour in the palette.
        Currently returns the first colour, but could be enhanced with
        clustering or frequency analysis.

        Returns:
            Colour: The dominant colour
        """
        return self._colours[0]


    def add_colour(self, colour: Colour) -> None:
        """
        Add a new colour to the palette.

        Args:
            colour (Colour): Colour to add
        """
        self._colours.append(colour)

    def remove_colour(self, colour: Colour) -> None:
        """
        Remove a colour from the palette.

        Args:
            colour (Colour): Colour to remove

        Raises:
            ValueError: If colour not in palette or would make palette empty
        """
        if colour not in self._colours:
            raise ValueError("Colour not in palette")
        if self.size <= 1:
            raise ValueError("Cannot remove last colour from palette")
        self._colours.remove(colour)


    @classmethod
    def from_list(cls, rgb_colours: List[Tuple[int, int, int]]) -> 'Palette':
        """
        Create a palette from a list of hex colour strings.

        Args:
            rgb_colours (List[tuples]): List of rgb tuples

        Returns:
            Palette: New Palette instance
        """
        colours = [Colour(rgb[0], rgb[1], rgb[2]) for rgb in rgb_colours]
        return cls(colours)

    def to_list(self) -> list[tuple[int, int, int]]:
        """
        Convert the palette to a list of rgb tuples.

        Returns:
            List[str]: List of hex colour codes
        """
        return [c.rgb for c in self._colours]

    @classmethod
    def from_hex_list(cls, hex_colours: List[str]) -> 'Palette':
        """
        Create a palette from a list of hex colour strings.

        Args:
            hex_colours (List[str]): List of hex colour codes

        Returns:
            Palette: New Palette instance
        """
        colours = [Colour.from_hex(hex_str) for hex_str in hex_colours]
        return cls(colours)

    def to_hex_list(self) -> List[str]:
        """
        Convert the palette to a list of hex colour strings.

        Returns:
            List[str]: List of hex colour codes
        """
        return [c.to_hex() for c in self._colours]


    @classmethod
    def from_string_list(cls, rgb_colours: List[str]) -> 'Palette':
        """
        Create a palette from a list of rgb colour strings.

        Args:
            rgb_colours (List[str]): List of rgb colour strings

        Returns:
            Palette: New Palette instance
        """
        colours = [Colour.from_string(rgb_str) for rgb_str in rgb_colours]
        return cls(colours)


    def to_string_list(self) -> List[str]:
        """
        Convert the palette to a list of rgb colour strings.

        Returns:
            List[str]: List of rgb colour strings
        """
        return [c.to_string() for c in self._colours]

    def __len__(self) -> int:
        """Get the number of colours in the palette."""
        return self.size

    def __getitem__(self, index: int) -> Colour:
        """Get a colour by index."""
        return self._colours[index]

    def __iter__(self):
        """Iterate over colours in the palette."""
        return iter(self._colours)

    def __repr__(self) -> str:
        """String representation of the palette."""
        hex_colours = [c.to_hex() for c in self._colours]
        return f"Palette(colours={hex_colours})"
